using System;
using UnityEngine;
using System.Collections;
using System.Collections.Generic;
using System.Text.RegularExpressions;
using SimpleJSON;
using System.Linq;
using MVR.FileManagementSecure;
using static ClockwiseSilver.Helpers;

namespace ClockwiseSilver
{
	public class ExpressionTool : MVRScript
	{
		private Atom her;
		private GenerateDAZMorphsControlUI morphControl;
		private JSONStorableBool isActive;		
		private JSONStorableFloat morphTime1JSON, morphSpeed1JSON, morphMax1JSON;
		private JSONStorableFloat morphTime2JSON, morphSpeed2JSON, morphMax2JSON;
		
		private JSONStorableString sString1;
		private UIDynamicTextField sText1;
		private List<string> sequence1 = new List<string>();
		private JSONStorableStringChooser jchooser1;
		private UIDynamicPopup popup1;
		
		private JSONStorableString sString2;
		private UIDynamicTextField sText2;
		private List<string> sequence2 = new List<string>();
		private JSONStorableStringChooser jchooser2;
		private UIDynamicPopup popup2;
		
		private bool stopping = false;
		private float timer1, timer2 = 0f;
		private int cIdx = 0;
		private int mIdx1, mIdx2 = -1;
		private int mmStart = 0;
		private int mmEnd = 1;
		private int opStart = 2;
		private int opEnd = 3;
		private List<DAZMorph> morphList = new List<DAZMorph>();
		private List<int> targetList = new List<int>();
		private List<float> currentList = new List<float>();
		
		//Credit to Physis for json save/load process
		//https://www.reddit.com/r/VAMscenes/comments/alhc6j/plugin_movement_helper/
		public static string saveExt = "exa";
		private string _lastBrowseDir;
		
		public override void Init() {
			try
			{
				her = containingAtom;
                JSONStorable geometry = her.GetStorableByID("geometry");
                DAZCharacterSelector character = geometry as DAZCharacterSelector;
                morphControl = character.morphsControlUI;
				List<string> choices = new List<string>();
				choices.Add("None");
			
				if (morphControl != null)
				{
					morphControl.GetMorphDisplayNames().ForEach((name) =>
					{
						DAZMorph morph = morphControl.GetMorphByDisplayName(name);
						if (morph.region.ToLower().Contains("express"))
						{
							choices.Add(name);
						}
					});
				}
				
				isActive = new JSONStorableBool ("isActive", false);
				RegisterBool (isActive);
				var toggle = CreateToggle(isActive);
				toggle.label = "Active";
				
				JSONStorableAction ToggleActive = new JSONStorableAction("Toggle Expressions", () =>
                {
                    isActive.val = !isActive.val;
                });
                RegisterAction(ToggleActive);
				
				var heading = CreateTextField(new JSONStorableString("header0",
					"Requires at least 3 expressions \nin each section.\n" +
					"Avoid using the same expression\nin both sections.\n" +
					"Avoid using formula expressions\nthat drive other parameters."));
				heading.height = 200f;
				
				//-----------------UI------------------
				
				morphTime1JSON = new JSONStorableFloat("Change Exp Time1", 2.5f, 0.05f, 5.0f, false);
				RegisterFloat(morphTime1JSON);
				CreateSlider(morphTime1JSON, false);
				
				morphSpeed1JSON = new JSONStorableFloat("Morph Speed1", 1.0f, 0.1f, 4.0f, false);
				RegisterFloat(morphSpeed1JSON);
				CreateSlider(morphSpeed1JSON, false);
				
				morphMax1JSON = new JSONStorableFloat("Expressions Max1", 0.4f, 0.05f, 1.0f, false);
				RegisterFloat(morphMax1JSON);
				CreateSlider(morphMax1JSON, false);
				
				morphTime2JSON = new JSONStorableFloat("Change Exp Time2", 1.4f, 0.05f, 5.0f, false);
				RegisterFloat(morphTime2JSON);
				CreateSlider(morphTime2JSON, false);
				
				morphSpeed2JSON = new JSONStorableFloat("Morph Speed2", 0.25f, 0.1f, 5.0f, false);
				RegisterFloat(morphSpeed2JSON);
				CreateSlider(morphSpeed2JSON, false);
				
				morphMax2JSON = new JSONStorableFloat("Expressions Max2", 0.4f, 0.05f, 1.0f, false);
				RegisterFloat(morphMax2JSON);
				CreateSlider(morphMax2JSON, false);
				
				var btn = CreateButton("Reset Expressions", false);
                btn.button.onClick.AddListener(() => { ResetExpressions(); });
				
				//------------------------------------Chooser 1------------------------------------
				sString1 = new JSONStorableString("Sequence1String", "");
				sString1.storeType = JSONStorableParam.StoreType.Full;
				RegisterString(sString1);
				sText1 = CreateTextField(sString1, true);
				sText1.height = 120;
				btn = CreateButton("Remove Last", true);
                btn.button.onClick.AddListener(() => { RemoveLastExpression1(); });
				btn.buttonColor = Color.gray;
				btn.textColor = Color.white;
				btn = CreateButton("Add Active", true);
				btn.buttonColor = Color.blue;
				btn.textColor = Color.white;
                btn.button.onClick.AddListener(() => { AddActive1(); });					
				
				jchooser1 = new JSONStorableStringChooser("Chooser1", choices, "None", "Express", ExpressionsCallback1);                
				RegisterStringChooser(jchooser1);
				popup1 = CreateFilterablePopup(jchooser1, true);
				popup1.popupPanelHeight = 440f;
				
				sString2 = new JSONStorableString("Sequence2String", "");
				sString2.storeType = JSONStorableParam.StoreType.Full;
				RegisterString(sString2);
				sText2 = CreateTextField(sString2, true);
				sText2.height = 120;
				btn = CreateButton("Remove Last", true);
                btn.button.onClick.AddListener(() => { RemoveLastExpression2(); });
				btn.buttonColor = Color.gray;
				btn.textColor = Color.white;				
				btn = CreateButton("Add Active", true);
				btn.buttonColor = Color.blue;
				btn.textColor = Color.white;
                btn.button.onClick.AddListener(() => { AddActive2(); });				
				
				jchooser2 = new JSONStorableStringChooser("Chooser2", choices, "None", "Express", ExpressionsCallback2);                
				RegisterStringChooser(jchooser2);
				popup2 = CreateFilterablePopup(jchooser2, true);
				popup2.popupPanelHeight = 440f;
				
				//------------------------Presets-------------------------
				CreateSpacer(true).height = 40;
				btn = CreateButton("Load Preset",true);
				btn.button.onClick.AddListener(() =>
				{
					SuperController.singleton.NormalizeMediaPath("Saves\\ExpToolPresets\\"); 
					SuperController.singleton.GetMediaPathDialog(LoadPreset, saveExt);
				});
				btn.buttonColor = Color.blue;
				btn.textColor = Color.white;

				btn = CreateButton("Save Preset", true);
				btn.button.onClick.AddListener(() =>
				{
					FileManagerSecure.CreateDirectory("Saves\\ExpToolPresets");
					SuperController.singleton.NormalizeMediaPath("Saves\\ExpToolPresets\\"); 
					SuperController.singleton.GetMediaPathDialog(SavePreset, saveExt);

					// Update the browser to be a Save browser
					uFileBrowser.FileBrowser browser = SuperController.singleton.mediaFileBrowserUI;
					browser.defaultPath =  "Saves\\ExpToolPresets";
					browser.SetTextEntry(true);
					browser.fileEntryField.text = String.Format("{0}.{1}", ((int)(DateTime.UtcNow - new DateTime(1970, 1, 1)).TotalSeconds).ToString(), saveExt);
					browser.ActivateFileNameField();
				});
				btn.buttonColor = Color.blue;
				btn.textColor = Color.white;
			}
			catch (Exception e)
			{
				SuperController.LogError("Exception caught: " + e);
			}
		}
		
		void Start()
        {	
			//Set Listeners
			isActive.toggle.onValueChanged.AddListener(checkVal =>
				{
					if (checkVal) { StartExpressions(); }
					else { StopExpressions(); }
				});
				
			ReloadExpressions();
			if (sequence1.Count == 0) { Preset1(); ReloadExpressions(); }
			if (sequence2.Count == 0) { Preset2(); ReloadExpressions(); }
			
			if (isActive.val && sequence1.Count != 0 && sequence2.Count != 0)
			{
				StartExpressions();
			}
        }
		
		private void AddActive1()
		{
			JSONStorable geometry = her.GetStorableByID("geometry");
			DAZCharacterSelector character = geometry as DAZCharacterSelector;
			morphControl = character.morphsControlUI;
		
			if (morphControl != null)
			{
				StopExpressions();
				sString1.val = "";
						
				morphControl.GetMorphDisplayNames().ForEach((name) =>
				{
					DAZMorph morph = morphControl.GetMorphByDisplayName(name);
					if (morph.region.ToLower().Contains("express") && morph.morphValue != 0f)
					{
						sequence1.Add(name);
					}
				});
				
				for (int i = 0; i < sequence1.Count; i++)
				{
					if (i > 0) { sString1.val += ",,"; }
					sString1.val += sequence1[i];
				}
				
				ReloadExpressions();
			}
		}
		
		private void AddActive2()
		{
			JSONStorable geometry = her.GetStorableByID("geometry");
			DAZCharacterSelector character = geometry as DAZCharacterSelector;
			morphControl = character.morphsControlUI;
		
			if (morphControl != null)
			{
				StopExpressions();
				sString2.val = "";
						
				morphControl.GetMorphDisplayNames().ForEach((name) =>
				{
					DAZMorph morph = morphControl.GetMorphByDisplayName(name);
					if (morph.region.ToLower().Contains("express") && morph.morphValue != 0f)
					{
						sequence2.Add(name);
					}
				});
				
				for (int i = 0; i < sequence2.Count; i++)
				{
					if (i > 0) { sString2.val += ",,"; }
					sString2.val += sequence2[i];
				}
				
				ReloadExpressions();
			}
		}
		
		void SavePreset(string presetPath)
        {
            if (String.IsNullOrEmpty(presetPath))
            {
                return;
            }
            _lastBrowseDir = presetPath.Substring(0, presetPath.LastIndexOfAny(new char[] { '/', '\\' })) + @"\";

            if ( !presetPath.ToLower().EndsWith(saveExt.ToLower()))
            {
                presetPath += "." + saveExt;
            }
            JSONClass saveJson = new JSONClass();
            saveJson["sequence1"] = new JSONArray();
            saveJson["sequence2"] = new JSONArray();
            foreach (string str in sequence1)
            {
				saveJson["sequence1"].Add(str);
            }
			foreach (string str in sequence2)
            {
				saveJson["sequence2"].Add(str);
            }
            this.SaveJSON(saveJson, presetPath);
        }

        void LoadPreset(string presetPath)
        {
            if (String.IsNullOrEmpty(presetPath))
            {
                return;
            }
			
			if (isActive.val) { isActive.val = false; }
			
            _lastBrowseDir = presetPath.Substring(0, presetPath.LastIndexOfAny(new char[] { '/', '\\' })) + @"\";
            var jsonFile = this.LoadJSON(presetPath);

			sString1.val = "";
			sequence1.Clear();
			foreach (JSONNode nd in jsonFile["sequence1"].AsArray)
            {
                sequence1.Add(nd.Value);
            }			
			for (int i = 0; i < sequence1.Count; i++)
			{
				if (i > 0) { sString1.val += ",,"; }
				sString1.val += sequence1[i];
			}
			
			sString2.val = "";
			sequence2.Clear();
			foreach (JSONNode nd in jsonFile["sequence2"].AsArray)
            {
                sequence2.Add(nd.Value);
            }			
			for (int i = 0; i < sequence2.Count; i++)
			{
				if (i > 0) { sString2.val += ",,"; }
				sString2.val += sequence2[i];
			}
			
			ReloadExpressions();
        }
		
		private void ReloadExpressions()
		{			
			sequence1.Clear();
			morphList.Clear();
			targetList.Clear();
			currentList.Clear();
			string[] splitString = Regex.Split(sString1.val, ",,");			
			foreach (string str in splitString)
			{
				if (str != "")
				{
					DAZMorph mph = morphControl.GetMorphByDisplayName(str);
					
					if (mph != null)
					{
						sequence1.Add(str);
						morphList.Add(mph);
						targetList.Add(0);
						currentList.Add(0f);
					}
				}
			}			
			mmStart = 0;
			mmEnd = morphList.Count;
			
			sequence2.Clear();
			splitString = Regex.Split(sString2.val, ",,");
			foreach (string str in splitString)
			{
				if (str != "")
				{
					DAZMorph mph = morphControl.GetMorphByDisplayName(str);
					
					if (mph != null)
					{
						sequence2.Add(str);
						morphList.Add(mph);
						targetList.Add(0);
						currentList.Add(0f);
					}
				}
			}	
			opStart = mmEnd;
			opEnd = morphList.Count;
		}
		
		private void ResetExpressions()
		{
			morphControl.GetMorphDisplayNames().ForEach((name) =>
                {
                    DAZMorph morph = morphControl.GetMorphByDisplayName(name);
                    if (morph.region.Contains("Expressions"))
					{
						morph.morphValue = 0.0f;
					}
                });
		}
		
		private void RemoveLastExpression1()
		{
			if(sequence1.Count == 0) { return; }			
			if (isActive.val) { isActive.val = false; }
			
			timer1 = -1.0f;
			timer2 = -1.0f;				
			ResetExpressions();			
			sequence1.RemoveAt(sequence1.Count - 1);
			sString1.val = "";
			for (int i = 0; i < sequence1.Count; i++)
			{
				if (i > 0) { sString1.val += ",,"; }
				sString1.val += sequence1[i];
			}
			
			ReloadExpressions();
		}
		
		private void RemoveLastExpression2()
		{
			if(sequence2.Count == 0) { return; }
			if (isActive.val) { isActive.val = false; }
			
			timer1 = -1.0f;
			timer2 = -1.0f;
			ResetExpressions();			
			sequence2.RemoveAt(sequence2.Count - 1);
			sString2.val = "";
			for (int i = 0; i < sequence2.Count; i++)
			{
				if (i > 0) { sString2.val += ",,"; }
				sString2.val += sequence2[i];
			}
			
			ReloadExpressions();
		}
		
		private void Preset1()
		{
			sString1.val = "";
			sequence1.Clear();
			
			sequence1.Add("Smile Full Face");
			sequence1.Add("Smile Open Full Face");
			sequence1.Add("Desire");
			sequence1.Add("Snarl Left");
			sequence1.Add("Flirting");
			
			for (int i = 0; i < sequence1.Count; i++)
			{
				if (i > 0) { sString1.val += ",,"; }
				sString1.val += sequence1[i];
			}
			
			ReloadExpressions();
		}
		
		private void Preset2()
		{
			sString2.val = "";
			sequence2.Clear();
			
			sequence2.Add("Concentrate");
			sequence2.Add("Glare");
			sequence2.Add("Shock");
			sequence2.Add("Surprise");
			
			for (int i = 0; i < sequence2.Count; i++)
			{
				if (i > 0) { sString2.val += ",,"; }
				sString2.val += sequence2[i];
			}
			
			ReloadExpressions();
		}
		
		private void ExpressionsCallback1(string s) {
			if (s == "None") { return; }
			if (sString1 != null)
			{
				if (isActive.val) { isActive.val = false; }
				sequence1.Add(s);				
				sString1.val = "";
				timer1 = -1.0f;
				timer2 = -1.0f;				
				ResetExpressions();
				
				for (int i = 0; i < sequence1.Count; i++)
				{
					if (i > 0) { sString1.val += ",,"; }
					sString1.val += sequence1[i];
				}
				
				ReloadExpressions();
			}
		}
		
		private void ExpressionsCallback2(string s) {
			if (s == "None") { return; }
			if (sString2 != null)
			{
				if (isActive.val) { isActive.val = false; }
				sequence2.Add(s);				
				sString2.val = "";
				timer1 = -1.0f;
				timer2 = -1.0f;				
				ResetExpressions();
				
				for (int i = 0; i < sequence2.Count; i++)
				{
					if (i > 0) { sString2.val += ",,"; }
					sString2.val += sequence2[i];
				}
				
				ReloadExpressions();
			}
		}
		
		private void SetMorphTargets1()
		{
			targetList[mIdx1] = 0;
			
			int randomIndex = UnityEngine.Random.Range(mmStart, mmEnd);
			if (targetList[randomIndex] != 0)
			{
				randomIndex++;
				if (randomIndex >= mmEnd)
				{
					randomIndex = mmStart;
				}
			}
			
			targetList[randomIndex] = 1;
			mIdx1 = randomIndex;			
			timer1 = morphTime1JSON.val * UnityEngine.Random.Range(0.5f, 1.5f);
		}
		
		private void SetMorphTargets2()
		{			
			targetList[mIdx2] = 0;
			
			int	randomIndex = UnityEngine.Random.Range(opStart, opEnd);
			
			if (targetList[randomIndex] != 0)
			{
				randomIndex++;
				if (randomIndex >= opEnd)
				{
					randomIndex = opStart;
				}
			}
			
			targetList[randomIndex] = 2;
			mIdx2 = randomIndex;
			timer2 = morphTime2JSON.val * UnityEngine.Random.Range(0.5f, 1.5f);
		}
		
		public void StartExpressions()
		{
			if (sequence1.Count < 3 && sequence2.Count < 3)
			{
				SuperController.LogMessage("SilverExpressionTool: At least 3 expressions required");
				isActive.val = false;
				return;
			}
			ResetExpressions();
			stopping = false;
			StartExpress1();
			StartExpress2();
		}
		
		public void StopExpressions()
		{
			stopping = true;
		}
		
		private void StartExpress1()
		{
			if (sequence1.Count < 3) { return; }
			
			int randomIndex = UnityEngine.Random.Range(mmStart, mmEnd);
			if (randomIndex == mIdx1)
			{
				randomIndex++;
				if (randomIndex >= mmEnd)
				{
					randomIndex = mmStart;
				}
			}
			mIdx1 = randomIndex;			
			float holdTime = morphTime1JSON.val * UnityEngine.Random.Range(0.75f, 1.25f);
			StartCoroutine(MorphIt(morphList[randomIndex], morphMax1JSON.val, morphSpeed1JSON.val, holdTime, true));
		}
		
		private void StartExpress2()
		{
			if (sequence2.Count < 3) { return; }
			
			int	randomIndex = UnityEngine.Random.Range(opStart, opEnd);			
			if (randomIndex == mIdx2)
			{
				randomIndex++;
				if (randomIndex >= opEnd)
				{
					randomIndex = opStart;
				}
			}
			mIdx2 = randomIndex;			
			float holdTime = morphTime2JSON.val * UnityEngine.Random.Range(0.75f, 1.25f);
			StartCoroutine(MorphIt(morphList[randomIndex], morphMax2JSON.val, morphSpeed2JSON.val, holdTime, false));
		}
		
		private IEnumerator MorphIt(DAZMorph morph, float max, float speed, float holdTime, bool isOne)
		{
			float val = 0;			
			while (isActive.val && val < 1f)
			{
				val += Time.deltaTime * speed;
				morph.morphValue = Mathf.Clamp(Helpers.QuadraticInOut(val) * max, 0f, 1f);
				yield return null;
			}
			
			val = 0;
			while (isActive.val && val < holdTime)
			{
				val += Time.deltaTime;
				yield return null;
			}
			
			//Start dialing up the new morph WHILE dialing down the current
			if (isActive.val)
			{
				if (isOne) { StartExpress1(); }
				else  { StartExpress2(); }
			}
			
			val = 1;			
			while (val > 0f)
			{
				val -= Time.deltaTime * speed;
				morph.morphValue = Mathf.Clamp(Helpers.QuadraticInOut(val) * max, 0f, 1f);
				yield return null;
			}
			
			//morph.morphValue = 0f;
		}

		private void OnDisable()
		{
			StopExpressions();
			ResetExpressions();
		}
		
		private void OnDestroy()
		{
			StopExpressions();
			ResetExpressions();
		}
	}
}